; These are equates, they work exactly like #define in C
; We use them here for the locations of the various data in VU memory
Transform		.equ  	0
LightTrans		.equ    4
Scale			.equ	8
NumVertsMem		.equ	8
LightDirs		.equ    9
LightCols		.equ	12
GIFTag			.equ    16
StartSTQ		.equ    17
StartNorm		.equ    18
StartVert		.equ	19

; This is a file full of macros that comes with VCL
.include "vcl_sml.i"

; Tell VCL that we want to be able to use all vf, and vi registers
.init_vf_all
.init_vi_all

.syntax new
.vu

; This is the entry point, and is where execution will start on VU1
--enter
--endenter

	fcset		0x000000					; VCL won't let us use CLIP without first zeroing
											; the clip flags

	iaddiu		iVert, vi00, 0				; Start vertex counter
	iaddiu		iVertPtr, vi00, 0			; Point to the first vert

	ilw.w		iNumVerts, NumVertsMem(vi00); Load the number of verts
	
	lq			fScales, Scale(vi00)		; A scale vector that we will add
											; to and scale the vertices by after
											; projecting them


	lq			fTransform[0], Transform+0(vi00)	; Load vertex transformation matrix
	lq			fTransform[1], Transform+1(vi00)
	lq			fTransform[2], Transform+2(vi00)
	lq			fTransform[3], Transform+3(vi00)
	
	lq			fLightTrans[0], LightTrans+0(vi00)	; Load Light transfromation matrix
	lq			fLightTrans[1], LightTrans+1(vi00)
	lq			fLightTrans[2], LightTrans+2(vi00)

    lq.xyz		fLightDirs[0], LightDirs+0(vi00)	; Load the light directions
    lq.xyz		fLightDirs[1], LightDirs+1(vi00)
    lq.xyz		fLightDirs[2], LightDirs+2(vi00)
    
    sub.xyz		fLightDirs[0], vf00, fLightDirs[0]	; Reverse Light directions
    sub.xyz		fLightDirs[1], vf00, fLightDirs[1]
    sub.xyz		fLightDirs[2], vf00, fLightDirs[2]

	lq.xyz		fLightCols[0], LightCols+0(vi00)	; Load the light colours
    lq.xyz		fLightCols[1], LightCols+1(vi00)
    lq.xyz		fLightCols[2], LightCols+2(vi00)
    lq.xyz		fAmbient,      LightCols+3(vi00)
    
    loi			128.0								; Load 128 for the alpha value
											
loop:										; For each vertex
	lq			Vert, StartVert(iVertPtr)	; Load the vector (currently in object space
											; and floating point format)
	
	; Transform the vertex by the projection matrix we passed in (this is a macro from
	; vcl_sml.i)
	MatrixMultiplyVertex Vert, fTransform, Vert
	
	; CLIPPING
	clipw.xyz	Vert, Vert					; This instruction checks if the vertex is outside
											; the viewing frustum. If it is, then the appropriate
											; clipping flags are set
	fcand		vi01, 0x3FFFF				; Bitwise AND the clipping flags with 0x3FFFF, this makes
											; sure that we get the clipping judgement for the last three
											; verts (i.e. that make up the triangle we are about to draw)
	iaddiu		iADC, vi01, 0x7FFF			; Add 0x7FFF. If any of the clipping flags were set this will
											; cause the triangle not to be drawn (any values above 0x8000
											; that are stored in the w component of XYZ2 will set the ADC
											; bit, which tells the GS not to perform a drawing kick on this
											; triangle.
											
	ilw.w		iNoDraw, StartSTQ(iVertPtr)	; Load the iNoDraw flag. If true we should 
	iadd		iADC, iADC, iNoDraw 		; set the ADC bit so the vert isn't drawn										
	 
	isw.w		iADC, StartVert(iVertPtr)	; Finally store the ADC bit back to memory
	; END CLIPPING
    
    ; TEXTURE COORDINATES
    div         q,    vf00[w], Vert[w]		; Work out 1.0f / w
    mul.xyz     Vert, Vert, q				; Project the vertex by dividing by W
    
    lq			STQ,  StartSTQ(iVertPtr)	; Load the texture coordinates
    mul			STQ,  STQ, q				; Divide by W
    sq			STQ,  StartSTQ(iVertPtr)	; Store the texture coordinates back
	; END TEXTURE COORDINATES

	;LIGHTING CALCULATIONS
	lq.xyz		Norm, StartNorm(iVertPtr)				; Load the normal
	
	mul.xyz     acc,  fLightTrans[0],  Norm[x]			; Transform by the rotation part of the world matrix
    madd.xyz    acc,  fLightTrans[1],  Norm[y]
    madd.xyz    Norm, fLightTrans[2],  Norm[z]

    mula.xyz	acc, fLightDirs[0], Norm[x]				; "Transform" the normal by the light direction matrix
    madd.xyz	acc, fLightDirs[1], Norm[y]				; This has the effect of outputting a vector with 
    madd.xyz	fIntensities, fLightDirs[2], Norm[z]	; three intensities, one for each light.
    mini.xyz	fIntensities, fIntensities, vf00[w]		; Clamp the intensity to 0..1
    max.xyz		fIntensities, fIntensities, vf00[x]
    
    mula.xyz	acc, fLightCols[0], fIntensities[x]		; Transform the intensities by the light colour matrix
    madda.xyz	acc, fLightCols[1], fIntensities[y]		; This gives the final total directional light colour
    madda.xyz	acc, fLightCols[2], fIntensities[z]
    madd.xyz	fIntensities, fAmbient, vf00[w]
    

	addi.w		fIntensities, vf00, i					; Store the  alpha value
    ftoi0		fIntensities, fIntensities				; Convert to int
    
	sq			fIntensities, StartNorm(iVertPtr)		; And write to the output buffer
	;END LIGHTING CALCULATIONS

	; VERTEX SCALING
    mula.xyz	acc, fScales, vf00[w]		; Move fScales into the accumulator (acc = fScales * 1.0)
 	madd.xyz	Vert, Vert, fScales			; Multiply and add the scales (Vert = Vert * fScales + fScales)
	ftoi4.xyz	Vert, Vert					; Convert the vertex to 12:4 fixed point for the GS

	sq.xyz		Vert, StartVert(iVertPtr)	; Write the vertex back to VU memory
	; END VERTEX SCALING
	
	iaddiu		iVert, iVert, 1				; Increment the vertex counter
	iaddiu		iVertPtr, iVertPtr, 3		; Increment the vertex pointer
	ibne		iVert, iNumVerts, loop		; Branch back to "loop" label if
											; iVert and iNumVerts are not equal
											
	iaddiu		iGIFTag, vi00, GIFTag		; Load the position of the giftag
	xgkick		iGIFTag						; and tell the PS2 to send that to the GS
--exit
--endexit